#include <msp430.h> 
#include <limits.h>
#include <stdint.h>

#define DMAANZ (42)
unsigned int tarfeld[DMAANZ];	// DMA-Ergebnisfeld fr die TACCR2-Werte.
uint16_t DHT11_temp,DHT11_humidity; // Messergebnisse des DHT11.

enum {DHT11_INAKTIVE,DHT11_AKTIVE,DHT11_BADCRC,DHT11_NOANSWER,DHT11_DATA} DHT11_zust;

#define SMCLOCK 8000000
#define DELAY18MS ((20e-3/8*SMCLOCK)+1)
#define BITTIME99us (unsigned int)(99e-6*SMCLOCK)

/*
 * Der DHT11 wird an P2.0 angeschlossen. Damit ist es ein Capture-Pin, dessen Ergebnisse per
 * DMA ausgelesen werden knnen.
 * Da der FG4618 keine Kalibrierdaten fr den DCO bereit stellt, erzeuge ich 8MHz Takt aus dem
 * Uhrenquarz an X2.
*/

/*
 * lese_DHT11: Anstossen des DHT, so dass dieser ein Datenpaket sendet.
 * lese_DHT11 gibt einen Lowpuls von 18ms aus, wie im Datenblatt des DHT11 beschrieben.
 * Am Ende der 18ms wird TIMERA1_ISR() aufgerufen.
 */
void lese_DHT11(void)
{
	if(DHT11_zust==DHT11_AKTIVE) return;
	DHT11_zust = DHT11_AKTIVE;
	/*
	 *  18ms lang Lowpegel auf der Datenleitung anlegen.
	 */
    TACTL = 0;					// Timer A2 anhalten
    P2DIR &= ~BIT0;				// whrend der Timereinstellung P2.0 hochohmig schalten
    P2SEL &= ~BIT0;
	TACCTL2 = OUTMOD_1 | CCIE;	// wollte per DMA treiben, nutze aber nur den Zeitgeber bis ISR()
	TACCR2 = DELAY18MS;
	TACTL = TASSEL_2 | ID_3 | TACLR;	// Teiler durch 8, damit 18ms erzeugbar sind.
	P2OUT &= ~BIT0;
	P2DIR |= BIT0;
	TACTL |= MC_2;				// Zhler loslaufen lassen -> TIMERA1_VECTOR
}

/*
 * TIMERA1_ISR(): Zweiter Schritt. Der Zeitgeber wird auf Capture-Betrieb umgestellt und der
 * DMA-Kanal gestartet. Am Ende der bertragung wird DMA_ISR() aufgerufen.
 */
#pragma vector=TIMERA1_VECTOR
__interrupt void TIMERA1_ISR(void)
{
	unsigned int vektor = TAIV;
	// Timer stoppen
	TACTL = 0;
	TACCTL2 = 0;
	_nop();
	/*
	 * Zwei Aufrufmglichkeiten:
	 * 1. TACCR2 compare: Die 18ms fr die Lowpegelausgabe sind abgelaufen.
	 * 2. TAIFG: Timeout fr die DHT11-Abfrage. Sptenstens nach 8192 Zyklen sollte die DHT11-Antwort empfangen sein.
	 * TAIFG tritt erst nach 65536 Zyklen auf.
	 */
	switch(vektor) {
	case TAIV_TAIFG :
						/*
						 * Keine Antwort des DHT11. Abbruch der Abfrage.
						 */
						DMACTL0 = 0;
						DMA0CTL = 0;
						DHT11_zust = DHT11_NOANSWER;
						_bic_SR_register_on_exit(SCG0|CPUOFF);
						break;
	case TAIV_TACCR2 :
						/*
						 * Nach den 18ms Lowpegel als Startimpuls, den Timer auf capture umstellen und die capture-Werte per
						 * DMA abholen.
						 * Timer auf Capture umstellen, DMA-Kanal einrichten, Portpin auf Eingang schalten und auf DMA-INT warten.
						 */
						// DMA-Kanal einrichten. Achtung: Das Triggersignal TACCR2 CCIFG darf noch nicht aktiv sein.
						//DMA0SA = (void (*)())&TACCR2;
						DMA0SAL = (unsigned int)&TACCR2;
						//DMA0DA = (void (*)())tarfeld;
						DMA0DAL = (unsigned int)&tarfeld[0];
						DMA0SZ = DMAANZ;
						DMACTL0 = DMA0TSEL_1;	// TACCR2_CCIFG als Triggersignal
						DMACTL1 = DMAONFETCH;
//						DMA0CTL = DMADT_0 | DMADSTINCR_3 | DMASRCINCR_0 | DMASWDW | DMALEVEL | DMAEN | DMAIE;
						// Portpin und Timer neu aufsetzen.
						P2DIR &= ~BIT0;
						P2SEL |= BIT0;
						TACCTL2 = CM_2 | CCIS_0 | SCS | CAP;// | CCIE; // IE knnte ein Fehler sein. ISR soll nicht aufgerufen werden.
						TACTL = TASSEL_2 | ID_0 | MC_2 | TAIE | TACLR;
						DMA0CTL = DMADT_0 | DMADSTINCR_3 | DMASRCINCR_0 | DMASWDW | DMAEN | DMAIE;
						/*
						 * Reihenfolge ist wichtig. TACTL vor DMACTL beschreiben.
						 */
						break;
	default : break;
	}
}

/*
 * DMA_ISR(): Das Datenpaket des DHT11 wurde komplett empfangen. Aus den Differenzen der Zeitstempel
 * werden die Bitwerte errechnet. Am Ende der Funktion muss ggf. der Prozessorkern wieder geweckt werden.
 */
#pragma vector=DMA_VECTOR
__interrupt void DMA_ISR(void)
{
	unsigned int lauf;
	uint16_t humidity;
	uint16_t temperature;
	uint8_t crc;
	/*
	 * Alles anhalten, Feld der Zeitdifferenzwerte auswerten, CPU starten, RETI.
	 */
	DMA0CTL = 0;
	TACTL = TACLR;
	TACCTL2 = 0;
	humidity = temperature = crc = 0;
	for(lauf=2;lauf<DMAANZ;lauf++) {
		uint16_t temp;
		humidity <<= 1;
		if(temperature&0x8000) humidity++;
		temperature <<= 1;
		if(crc&0x80) temperature++;
		crc <<= 1;
		temp = tarfeld[lauf]-tarfeld[lauf-1];
		if(temp>(unsigned int)INT_MAX+1) temp = -temp; //(~temp)+1;
		if(temp>BITTIME99us) crc++;
	}
	lauf = ((humidity>>8)+(humidity&0xff)+(temperature>>8)+(temperature&0xff))&0xff;
	DHT11_zust = (lauf==crc)?DHT11_DATA:DHT11_BADCRC;
	DHT11_temp = temperature;
	DHT11_humidity = humidity;
	_bic_SR_register_on_exit(SCG0|CPUOFF);
}

/*
 * main.c
 */
int main(void) {
    WDTCTL = WDTPW | WDTHOLD;	// Stop watchdog timer
    FLL_CTL0 |= DCOPLUS + XCAP18PF;           // DCO+ set, freq = xtal x D x N+1
    SCFI0 |= FN_4;                            // x2 DCO freq, 8MHz nominal DCO
    SCFQCTL = 121;                            // (121+1) x 32768 x 2 = 7.99 MHz
    _enable_interrupts();
    do {
    	lese_DHT11();
    	_bis_SR_register(CPUOFF);
    	_nop();
    } while(DHT11_zust != DHT11_DATA);
    _nop();	// Breakpoint
	return 0;
}
